/*!
 *****************************************************************************
 *
 * @File       actions.c
 * ---------------------------------------------------------------------------
 *
 * Copyright (c) Imagination Technologies Ltd.
 * 
 * The contents of this file are subject to the MIT license as set out below.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
 * THE SOFTWARE.
 * 
 * Alternatively, the contents of this file may be used under the terms of the 
 * GNU General Public License Version 2 ("GPL")in which case the provisions of
 * GPL are applicable instead of those above. 
 * 
 * If you wish to allow use of your version of this file only under the terms 
 * of GPL, and not to allow others to use your version of this file under the 
 * terms of the MIT license, indicate your decision by deleting the provisions 
 * above and replace them with the notice and other provisions required by GPL 
 * as set out in the file called "GPLHEADER" included in this distribution. If 
 * you do not delete the provisions above, a recipient may use your version of 
 * this file under the terms of either the MIT license or GPL.
 * 
 * This License is also included in this distribution in the file called 
 * "MIT_COPYING".
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <uapi/vxd.h>

#include "vxd_util.h"
#include "helpers.h"


static struct vxd_util_ctx vxd_util_ctxs[VXD_MAX_CTXS];
static int cur_vxd_ctx;

/* INIT */
void vxd_util_init()
{
	int i;
	for (i = 0; i < VXD_MAX_CTXS; i++) {
		struct vxd_util_ctx *ctx = &vxd_util_ctxs[i];
		int j, str_count = sizeof(ctx->str_ids)/sizeof(ctx->str_ids[0]);
		int buf_count = sizeof(ctx->bufs)/sizeof(struct vxd_buffer);
		ctx->fd = VXD_INVALID_ID;
		ctx->heap_id = VXD_INVALID_ID;
		ctx->cur_str_id = 1;
		ctx->cur_buf_id = 1;
		for (j = 0; j < str_count; j++)
			ctx->str_ids[j] = VXD_INVALID_ID;
		for (j = 0; j < buf_count; j++)
			ctx->bufs[j].id = VXD_INVALID_ID;
	}
	cur_vxd_ctx = 0;
}

static int vxd_get_free_str_slot(struct vxd_util_ctx *ctx)
{
	int i, str_count = sizeof(ctx->str_ids)/sizeof(ctx->str_ids[0]);

	for (i = 1; i < str_count; i++)
		if (ctx->str_ids[i] == VXD_INVALID_ID)
			return i;

	return VXD_INVALID_ID;	
}


/* OPEN */
void vxd_action_open_help(const char *name)
{
	(void)name;
	printf("\tOpen link to vxd\n");
}

int vxd_action_open_exec(const char *arg)
{
	int fd;
	(void)arg;

	vxd_dbg("  OPEN: ctx: %d\n", cur_vxd_ctx);

	fd = vxd_open(0);
	if (fd < 0) {
		vxd_err("  failed to open device: %d!\n", fd);
		return -1;
	}
	vxd_util_ctxs[cur_vxd_ctx].fd = fd;
	vxd_dbg("  OPEN OK: ctx: %d, fd: %d\n", cur_vxd_ctx, fd);
	return 0;
}

/* CLOSE */
void vxd_action_close_help(const char *name)
{
	(void)name;
	printf("\tClose link to vxd\n");
}

int vxd_action_close_exec(const char *arg)
{
	int fd = vxd_util_ctxs[cur_vxd_ctx].fd, ret;
	(void)arg;

	vxd_dbg("  CLOSE: ctx: %d, fd: %d\n", cur_vxd_ctx, fd);

	ret = vxd_close(fd);
	if (ret) {
		vxd_dbg("  CLOSE FAIL, ret: %d\n", ret);
	} else {
		vxd_dbg("  CLOSE OK\n");
		vxd_util_ctxs[cur_vxd_ctx].fd = -1;
	}

	return ret;
}

/* GET CORE PROPERTIRES */
void vxd_action_get_props_help(const char *name)
{
	(void)name;
	printf("\tfetch and print core properties\n");
}

const char *vxd_translate_chroma_fmt(unsigned int val)
{
	switch (val) {
		case 1:
			return "4:2:0";
		case 2:
			return "4:2:2";
		case 3:
			return "4:4:4";
		default:
			return "unknown";
	}
}

int vxd_action_get_props_exec(const char *arg)
{
	struct vxd_core_props props;
	struct vxd_util_ctx *ctx = &vxd_util_ctxs[cur_vxd_ctx];
	(void)arg;
	int ret;

	vxd_dbg("  GET PROPS: ctx: %d, fd: %d\n", cur_vxd_ctx, ctx->fd);

	ret = vxd_ioctl(ctx->fd, VXD_IOCTL_PROPS, (void *)&props);
	if (ret) {
		vxd_err("  GET PROPS FAIL: ret: %d\n", ret);
	} else {
		unsigned int i;

		ctx->heap_id = props.internal_heap_id;

		vxd_dbg("  GET PROPS: OK\n");
		vxd_dbg("  GET PROPS: id: 0x%08x\n", props.id);
		vxd_dbg("  GET PROPS: core_rev: 0x%08x\n", props.core_rev);
		vxd_dbg("  GET PROPS: pvdec_core_id: 0x%08x\n",
				props.pvdec_core_id);
		vxd_dbg("  GET PROPS: mmu_config0: 0x%08x\n", props.mmu_config0);
		vxd_dbg("  GET PROPS: mmu_config1: 0x%08x\n", props.mmu_config1);
		vxd_dbg("  GET PROPS: mtx_ram_size: 0x%08x\n", props.mtx_ram_size);
		vxd_dbg("  GET PROPS: pixel_max_frame_cfg: 0x%08x\n",
				props.pixel_max_frame_cfg);
		vxd_dbg("  GET PROPS: MTX RAM: %u\n", props.mtx_ram_size);
		vxd_dbg("  GET PROPS: num pixel pipes: %u\n", VXD_NUM_PIX_PIPES(props));
		vxd_dbg("  GET PROPS: num entropy pipes: %u\n", VXD_NUM_ENT_PIPES(props));
		vxd_dbg("  GET PROPS: internal heap id: %d\n", props.internal_heap_id);

		for (i = 0; i < VXD_NUM_PIX_PIPES(props); i++) {
			vxd_dbg("  GET PROPS: pix pipe %u, 0x%08x, misc: 0x%08x\n",
					i, props.pixel_pipe_cfg[i],
					props.pixel_misc_cfg[i]);
		}
		vxd_dbg("  GET PROPS:\n");
		vxd_dbg("  GET PROPS: PARSED FEATURES:\n");
		vxd_dbg("  GET PROPS:   CORE REV: %u.%u.%u\n", VXD_MAJ_REV(props),
				VXD_MIN_REV(props), VXD_MAINT_REV(props));
		vxd_dbg("  GET PROPS:   MMU BUS WIDTH: %u\n",
				VXD_EXTRN_ADDR_WIDTH(props));
		vxd_dbg("  GET PROPS:   MAX FRAME MSVDX: %ux%u\n",
				VXD_MAX_WIDTH_MSVDX(props),
				VXD_MAX_HEIGHT_MSVDX(props));
		vxd_dbg("  GET PROPS:   MAX FRAME PVDEC: %ux%u\n",
				VXD_MAX_WIDTH_PVDEC(props),
				VXD_MAX_HEIGHT_PVDEC(props));
		for (i = 0; i < VXD_NUM_PIX_PIPES(props); i++) {
			vxd_dbg("  GET PROPS: PIPE %u\n", i);
			vxd_dbg("  GET PROPS:   MPEG2: %u\n",
					VXD_HAS_MPEG2(props, i));
			vxd_dbg("  GET PROPS:   MPEG4: %u\n",
					VXD_HAS_MPEG4(props, i));
			vxd_dbg("  GET PROPS:   H264: %u\n",
					VXD_HAS_H264(props, i));
			vxd_dbg("  GET PROPS:   VC1: %u\n",
					VXD_HAS_VC1(props, i));
			vxd_dbg("  GET PROPS:   WMV9: %u\n",
					VXD_HAS_WMV9(props, i));
			vxd_dbg("  GET PROPS:   JPEG: %u\n",
					VXD_HAS_JPEG(props, i));
			vxd_dbg("  GET PROPS:   MPEG4 data partition: %u\n",
					VXD_HAS_MPEG4_DATA_PART(props, i));
			vxd_dbg("  GET PROPS:   AVS: %u\n",
					VXD_HAS_AVS(props, i));
			vxd_dbg("  GET PROPS:   REAL: %u\n",
					VXD_HAS_REAL(props, i));
			vxd_dbg("  GET PROPS:   VP6: %u\n",
					VXD_HAS_VP6(props, i));
			vxd_dbg("  GET PROPS:   VP8: %u\n",
					VXD_HAS_VP8(props, i));
			vxd_dbg("  GET PROPS:   SORENSON: %u\n",
					VXD_HAS_SORENSON(props, i));
			vxd_dbg("  GET PROPS:   HEVC 64K STILL: %u\n",
					VXD_HAS_HEVC_64K_STILL(props, i));
			vxd_dbg("  GET PROPS:   PIPE_ID: %u\n",
					VXD_PIX_PIPE_ID(props, i));
			vxd_dbg("  GET PROPS:   NUM STREAMS: %u\n",
					VXD_PIX_NUM_STRS(props, i));
			vxd_dbg("  GET PROPS:   SCALING: %u\n",
					VXD_HAS_SCALING(props, i));
			vxd_dbg("  GET PROPS:   ROTATION: %u\n",
					VXD_HAS_ROTATION(props, i));
			vxd_dbg("  GET PROPS:   HEVC RANGE EXTENSIONS: %u\n",
					VXD_HAS_HEVC_REXT(props, i));
			vxd_dbg("  GET PROPS:   MAX BIT DEPTH: %u\n",
					VXD_MAX_BIT_DEPTH(props, i));
			vxd_dbg("  GET PROPS:   HEVC MAX CHROMAT FORMAT: %s\n",
					vxd_translate_chroma_fmt(
						VXD_MAX_HEVC_CHROMA_FMT(props, i)));
			vxd_dbg("  GET PROPS:   H264 MAX CHROMAT FORMAT: %s\n",
					vxd_translate_chroma_fmt(
						VXD_MAX_H264_CHROMA_FMT(props, i)));
		}
	}

	return ret;
}

/* STREAM CREATE */
void vxd_action_str_creat_help(const char *name)
{
	(void)name;
	printf("\tSend ioctl to create a stream\n"
		"\tEach stream is assigned with a unique id internally,"
		" which start from 1 and is incremented for each subsequent"
		" buffer.\n");
}

int vxd_action_str_creat_exec(const char * arg)
{
	struct vxd_util_ctx *ctx = &vxd_util_ctxs[cur_vxd_ctx];
	int ret;
	struct vxd_create_stream_data data;
	const char *args = strchr(arg, ' '); /* skip action code */
	uint32_t *params;
	uint8_t *none;
	uint32_t str_type;
	size_t params_count = vxd_parse_ints(args, 0, &params, &none);

	if (params_count == 1)
		str_type = params[0];
	else
		str_type = VXD_STR_TYPE_LOOPBACK;

	ctx->cur_str_id = vxd_get_free_str_slot(ctx);
	if (ctx->cur_str_id == VXD_INVALID_ID) {
		vxd_err("  STREAM CREATE FAIL: failed to find free slot!\n");
		ret = 1;
		goto out;
	}

	data.stream_type = str_type;
	vxd_dbg("  STREAM CREATE: ctx: %d, fd: %d, str: %d type: %d\n",
			cur_vxd_ctx, ctx->fd, ctx->cur_str_id, str_type);

	ret = vxd_ioctl(ctx->fd, VXD_IOCTL_STREAM_CREATE, (void *)&data);
	VXD_CUR_STR_ID(ctx) = data.stream_id;
	if (ret)
		vxd_err("  STREAM CREATE FAIL: ret: %d\n", ret);
	else
		vxd_dbg("  STREAM CREATE OK, new internal id: %d, driver id: %d)\n",
				ctx->cur_str_id, VXD_CUR_STR_ID(ctx));

out:
	free(params);
	free(none);

	return ret;
}

/* STREAM SWITCH */
void vxd_action_str_switch_help(const char *name)
{
	printf("\tSwitch stream context.\n"
		"\t  --%s=<str context id>\n"
		"\tE.g. to change the context to the first created stream:\n"
		"\t  --%s=1\n", name, name);
}

int vxd_action_str_switch_exec(const char * arg)
{
	struct vxd_util_ctx *ctx = &vxd_util_ctxs[cur_vxd_ctx];
	int ret = 0;
	const char *args = strchr(arg, ' '); /* skip action code */
	uint32_t *params;
	uint8_t *none;
	size_t params_count = vxd_parse_ints(args, 0, &params, &none);
	int str_id, str_count = sizeof(ctx->str_ids)/sizeof(ctx->str_ids[0]);

	if (params_count < 1) {
		vxd_err("  STREAM SWITCH FAIL: stream id missing!\n");
		ret = 1;
		goto out;
	}

	str_id = params[0];

	if (str_id > str_count || str_id < 0) {
		vxd_err("  STREAM SWITCH FAIL: invalid stream id (%d)!\n", str_id);
		ret = 1;
		goto out;
	}

	if (ctx->str_ids[str_id] == VXD_INVALID_ID) {
		vxd_err("  STREAM SWITCH FAIL: trying to switch to invalid ctx (%d)!\n", str_id);
		ret = 1;
		goto out;
	}

	ctx->cur_str_id = str_id;

	vxd_dbg("  STREAM SWITCH OK, internal stream id: %d, driver %d\n",
			ctx->cur_str_id, VXD_CUR_STR_ID(ctx));

out:
	free(params);
	free(none);

	return ret;
}

/* STREAM DESTROY */
void vxd_action_str_destr_help(const char *name)
{
	(void)name;
	printf("\tSend ioctl to destroy the current stream\n");
}

int vxd_action_str_destr_exec(const char * arg)
{
	struct vxd_util_ctx *ctx = &vxd_util_ctxs[cur_vxd_ctx];
	int ret;
	struct vxd_destroy_stream_data data = {
		.stream_id = VXD_CUR_STR_ID(ctx),
	};
	(void)arg;

	vxd_dbg("  STREAM DESTROY: ctx: %d, fd: %d, str: %d, vxd str id: %d\n",
			cur_vxd_ctx, ctx->fd, ctx->cur_str_id, data.stream_id);

	ret = vxd_ioctl(ctx->fd, VXD_IOCTL_STREAM_DESTROY, (void *)&data);
	if (ret)
		vxd_err("  STREAM DESTROY FAIL: ret: %d\n", ret);
	else
		vxd_dbg("  STREAM DESTROY OK\n");

	return ret;
}

/* SEND MSG */
void vxd_action_send_msg_help(const char *name)
{
	printf("\tSend message, payload specified in 32bits words,"
			" in hex, e.g.:\n"
			"\t  --%s=\"0x1 0x10203040 200\"\n"
			"\tsends following message:\n"
			"\t  <current stream id>\n"
			"\t  <flags = 0>\n"
			"\t  <payload size>\n"
			"\t  0x00000001\n"
			"\t  0x10203040\n"
			"\t  0x00000200\n", name);
}

int vxd_action_send_msg_exec(const char * arg)
{
	struct vxd_util_ctx *ctx = &vxd_util_ctxs[cur_vxd_ctx];
	int ret;
	int str_id = VXD_CUR_STR_ID(ctx);
	size_t msg_size = 3;
	const char *args = strchr(arg, ' '); /* skip action code */
	uint32_t *msg;
	uint8_t *none;
	unsigned int i;
	int num_ints = vxd_parse_ints(args, msg_size, &msg, &none);
	msg[0] = str_id; /* stream id */
	msg[1] = 0; /* out flags */
	msg[2] = sizeof(uint32_t)*num_ints; /* payload size */
	msg_size += num_ints;

	vxd_dbg("  SEND MSG: ctx: %d, fd: %d, str: %d, arg: %s\n",
			cur_vxd_ctx, ctx->fd, ctx->cur_str_id, arg);
	vxd_dbg("  SEND MSG: vxd stream id: %d\n",
			VXD_CUR_STR_ID(ctx));

	for (i = 0; i < msg_size; i++)
		vxd_dbg("  SEND MSG:    [%03d] 0x%08x\n", i, msg[i]);


	ret = vxd_send_msg(ctx->fd, msg, msg_size);
	if (ret)
		vxd_err("  SEND MSG FAIL: ret: %d\n", ret);
	else
		vxd_dbg("  SEND MSG OK\n");

	free(msg);
	free(none);

	return ret;
}

/* RECV MSG */
void vxd_action_recv_msg_help(const char *name)
{
	printf("\tReceive message, optional pattern (32bits words in hex)"
		" to verify the message can be specified, e.g.:\n"
		"\t  --%s=\"0x2 200 N N 0x10203040 N\"\n"
		"\twill verify that the message matches following pattern\n"
		"\t  0x00000002\n"
		"\t  0x00000200\n"
		"\t  N - ignored\n"
		"\t  N - ignored\n"
		"\t  0x10203040\n"
		"\t  N - ignored\n",
		name);
}

#define RECV_BUF_SIZE 0x1000

int vxd_action_recv_msg_exec(const char * arg)
{
	struct vxd_util_ctx *ctx = &vxd_util_ctxs[cur_vxd_ctx];
	uint32_t buf[RECV_BUF_SIZE] = { 0 };
	unsigned int i;
	int count, ret = 0, ignore_flags = 0;
	const char *args = strchr(arg, ' '); /* skip action code */
	uint32_t *msg;
	uint8_t *none;
	size_t msg_size = vxd_parse_ints(args, 0, &msg, &none);
	size_t recv_msg_size;
	struct vxd_fw_msg *vxd_msg = NULL;

	vxd_dbg("  RECV MSG: ctx: %d, fd: %d, str: %d\n",
			cur_vxd_ctx, ctx->fd, ctx->cur_str_id);

	for (i = 0; i < msg_size; i++)
		if (none[i])
			vxd_dbg("  RECV MSG: pattern [%03d] none\n", i);
		else
			vxd_dbg("  RECV MSG: pattern [%03d] 0x%08x\n",
				i, msg[i]);

	count = vxd_recv_msg(ctx->fd, buf, RECV_BUF_SIZE);
	if (count < 0) {
		vxd_err("failed to read message!\n");
		ret = -1;
		goto out;
	}

	recv_msg_size = count/sizeof(uint32_t);

	if (msg_size > recv_msg_size) {
		vxd_err("received message is too small!\n");
		ret = -1;
		goto out;
	}

	if (msg_size)
		vxd_dbg("  RECV MSG: [ id] %10s - [got]\n", "[expected]");
	for (i = 0; i < msg_size; i++) {
		if (none[i]) {
			vxd_dbg("  RECV MSG: [%03d] %10s - 0x%08x\n",
				i, "none", buf[i]);
		} else {
			vxd_dbg("  RECV MSG: [%03d] 0x%08x - 0x%08x\n",
				i, msg[i], buf[i]);
			if (msg[i] != buf[i]) {
				vxd_err("mismatch in message!\n");
				ret = -1;
				goto out;
			}
		}
	}

	if (msg_size && msg_size < recv_msg_size)
		vxd_dbg("  RECV MSG: received message bigger than expected!\n");

	for (;i < recv_msg_size; i++)
		vxd_dbg("  RECV MSG: [%03d] 0x%08x\n", i, buf[i]);

	if (recv_msg_size < 3) {
		vxd_dbg("  RECV MSG: header in received message too small!\n");
		ret = -1;
		goto out;
	}

	vxd_msg = (struct vxd_fw_msg *)buf;

	/* Check if output flags were specified explicitly in expected output */
	if (msg_size >= 2) {
		ignore_flags = 1;
	}

	/* Check the output flags in received message if output flags
	 * weren't specified explicitly in expected output */
	if (!ignore_flags && vxd_msg->out_flags) {
		vxd_dbg("  RECV MSG: flags [%08x] indicate failure!\n",
				vxd_msg->out_flags);
		ret = -1;
		goto out;
	}

out:
	free(msg);
	free(none);

	if (ret)
		vxd_err("  RECV MSG FAIL: ret: %d\n", ret);
	else
		vxd_dbg("  RECV MSG OK\n");

	return ret;
}

/* ALLOCATE A BUFFER */
void vxd_action_alloc_help(const char *name)
{
	printf("\tAllocate a buffer of a size specified in bytes (in hex).\n"
		"\t  --%s=<buffer size in bytes>\n"
		"\tE.g. to allocate buffer 1000 (0x3e8) bytes large buffer:\n"
		"\t  --%s=3e8\n"
		"\tEach buffer is assigned with a unique id internally,"
		" which start from 1 and is incremented for each subsequent"
		" buffer.\n"
		"\tThis id can be used to identify buffer in other commands,"
		" e.g. to map it.\n", name, name);
}

int vxd_action_alloc_exec(const char * arg)
{
	struct vxd_util_ctx *ctx = &vxd_util_ctxs[cur_vxd_ctx];
	struct vxd_alloc_data data = { 0 };
	int ret;
	const char *args = strchr(arg, ' '); /* skip action code */
	uint32_t *params;
	uint8_t *none;
	size_t params_count = vxd_parse_ints(args, 0, &params, &none);
	size_t buf_size;

	vxd_dbg("  ALLOC: ctx: %d, fd: %d\n", cur_vxd_ctx, ctx->fd);

	if (ctx->cur_buf_id >= (int)(sizeof(ctx->bufs)/sizeof(ctx->bufs[0]))) {
		vxd_err("  ALLOC FAIL: all buffers slots (%lu) used!\n",
				(long unsigned int)sizeof(ctx->bufs)/sizeof(ctx->bufs[0]));
		ret = 1;
		goto out;
	}

	if (ctx->heap_id == VXD_INVALID_ID) {
		vxd_err("  ALLOC FAIL: invalid heap id!\n");
		ret = 1;
		goto out;
	}

	if (params_count < 1) {
		vxd_err("  ALLOC FAIL: buffer size missing!\n");
		ret = 1;
		goto out;
	}

	buf_size = params[0];

	vxd_dbg("  ALLOC: heap: %d, size: %zu\n", ctx->heap_id, buf_size);

	data.heap_id = ctx->heap_id;
	data.size = buf_size;
	data.attributes = 0;

	ret = vxd_ioctl(ctx->fd, VXD_IOCTL_ALLOC, (void *)&data);
	if (ret) {
		vxd_err("  ALLOC IOCTL FAIL: %d\n", ret);
		goto out;
	}
	ctx->bufs[ctx->cur_buf_id].id = data.buf_id;
	ctx->bufs[ctx->cur_buf_id].size = buf_size;
	ctx->cur_buf_id++;

out:
	free(params);
	free(none);

	if (ret)
		vxd_err("  ALLOC FAIL: ret: %d\n", ret);
	else
		vxd_dbg("  ALLOC OK, driver buf id: %d\n", data.buf_id);

	return ret;
}

/* FREE A BUFFER */
void vxd_action_free_help(const char *name)
{
	printf("\tFree the buffer.\n"
		"\t  --%s=<buf_id>\n"
		"\tE.g. to free the third allocated buffer:\n"
		"\t  --%s=3\n", name, name);
}

int vxd_action_free_exec(const char * arg)
{
	struct vxd_util_ctx *ctx = &vxd_util_ctxs[cur_vxd_ctx];
	struct vxd_free_data data = { 0 };
	int ret;
	const char *args = strchr(arg, ' '); /* skip action code */
	uint32_t *params, buf_id;
	uint8_t *none;
	size_t params_count = vxd_parse_ints(args, 0, &params, &none);

	vxd_dbg("  FREE: ctx: %d, fd: %d\n", cur_vxd_ctx, ctx->fd);

	if (params_count < 1) {
		vxd_err("  FREE FAIL: buffer id missing!\n");
		ret = 1;
		goto out;
	}

	buf_id = params[0];
	vxd_dbg("  FREE: buf id: %d\n", buf_id);

	if (buf_id >= sizeof(ctx->bufs)/sizeof(ctx->bufs[0])) {
		vxd_err("  FREE FAIL: buffer id (%d) too large!\n", buf_id);
		ret = 1;
		goto out;
	}

	if (ctx->bufs[buf_id].id == VXD_INVALID_ID) {
		vxd_err("  FREE FAIL: invalid buffer id!\n");
		ret = 1;
		goto out;
	}

	vxd_dbg("  FREE: driver buf id: %d, size: %zu, virt: 0x%08x\n",
			ctx->bufs[buf_id].id, ctx->bufs[buf_id].size,
			ctx->bufs[buf_id].virt_addr);

	data.buf_id = ctx->bufs[buf_id].id;

	ret = vxd_ioctl(ctx->fd, VXD_IOCTL_FREE, (void *)&data);
	if (ret) {
		vxd_err("  FREE IOCTL FAIL: %d\n", ret);
		goto out;
	}

	ctx->bufs[buf_id].id = VXD_INVALID_ID;
	ctx->bufs[buf_id].size = 0;
	ctx->bufs[buf_id].virt_addr = 0;

out:
	free(params);
	free(none);

	if (ret)
		vxd_err("  FREE FAIL: ret: %d\n", ret);
	else
		vxd_dbg("  FREE OK, driver buf id: %d\n", data.buf_id);

	return ret;
}

/* MAP THE BUFFER INTO VXD MMU */
void vxd_action_vxd_map_help(const char *name)
{
	printf("\tMap a buffer into VXD's stream PTD.\n"
		"\t --%s=\"<buf_id> <virtual_address>\"\n"
		"\tE.g. to map buffer with id 1 (so the first one that"
		" was allocated, as buffers ids are simply incremented)"
		" to virtual address 0x30000000:\n"
		"\t --%s=\"1 30000000\"\n", name, name);
}

int vxd_action_vxd_map_exec(const char * arg)
{
	struct vxd_util_ctx *ctx = &vxd_util_ctxs[cur_vxd_ctx];
	int ret;
	const char *args = strchr(arg, ' '); /* skip action code */
	uint32_t *params, buf_id, virt_addr, flags;
	uint8_t *none;
	size_t params_count = vxd_parse_ints(args, 0, &params, &none);
	struct vxd_map_data data = { 0 };
	int str_id = VXD_CUR_STR_ID(ctx);

	vxd_dbg("  VXD MAP: ctx: %d, fd: %d\n", cur_vxd_ctx, ctx->fd);

	if (params_count < 1) {
		vxd_err("  VXD MAP FAIL: buffer id missing!\n");
		ret = 1;
		goto out;
	}

	if (params_count < 2) {
		vxd_err("  VXD MAP FAIL: virt addr missing!\n");
		ret = 1;
		goto out;
	}

	buf_id = params[0];
	virt_addr = params[1];
	if (params_count == 3)
		flags = params[2];
	else
		flags = VXD_MAP_FLAG_NONE;
	vxd_dbg("  VXD MAP: buf id: %d, virt_addr: 0x%08x flags: 0x%x\n",
			buf_id, virt_addr, flags);

	if (buf_id >= sizeof(ctx->bufs)/sizeof(ctx->bufs[0])) {
		vxd_err("  VXD MAP FAIL: buffer id (%d) too large!\n", buf_id);
		ret = 1;
		goto out;
	}

	if (ctx->bufs[buf_id].id == VXD_INVALID_ID) {
		vxd_err("  VXD MAP FAIL: invalid buffer id!\n");
		ret = 1;
		goto out;
	}

	vxd_dbg("  VXD MAP: driver buf id: %d, size: %zu\n", ctx->bufs[buf_id].id,
			ctx->bufs[buf_id].size);

	data.stream_id = str_id;
	data.buf_id = ctx->bufs[buf_id].id;
	data.virt_addr = virt_addr;
	data.flags = flags;

	ret = vxd_ioctl(ctx->fd, VXD_IOCTL_VXD_MAP, (void *)&data);
	if (ret) {
		vxd_err("  VXD MAP IOCTL FAIL: %d\n", ret);
		goto out;
	}

	vxd_dbg("  VXD MAP: vxd virt addr: 0x%08x\n", data.virt_addr);

	ctx->bufs[buf_id].virt_addr = data.virt_addr;

out:
	free(params);
	free(none);

	if (ret)
		vxd_err("  VXD MAP FAIL: ret: %d\n", ret);
	else
		vxd_dbg("  VXD MAP OK\n");

	return ret;
}

/* UNMAP THE BUFFER FROM VXD MMU */
void vxd_action_vxd_unmap_help(const char *name)
{
	printf("\tUnmap the buffer from VXD's stream PTD.\n"
		"\t --%s=\"<buf_id>\"\n"
		"\tE.g. to unmap buffer with id 1 (so the first one that"
		" was allocated, as buffers ids are simply incremented):\n"
		"\t --%s=\"1\"\n", name, name);
}

int vxd_action_vxd_unmap_exec(const char * arg)
{
	struct vxd_util_ctx *ctx = &vxd_util_ctxs[cur_vxd_ctx];
	int ret;
	const char *args = strchr(arg, ' '); /* skip action code */
	uint32_t *params, buf_id;
	uint8_t *none;
	size_t params_count = vxd_parse_ints(args, 0, &params, &none);
	struct vxd_unmap_data data = { 0 };
	int str_id = VXD_CUR_STR_ID(ctx);

	vxd_dbg("  VXD UNMAP: ctx: %d, fd: %d\n", cur_vxd_ctx, ctx->fd);

	if (params_count < 1) {
		vxd_err("  VXD UNMAP FAIL: buffer id missing!\n");
		ret = 1;
		goto out;
	}


	buf_id = params[0];
	vxd_dbg("  VXD UNMAP: buf id: %d\n", buf_id);

	if (buf_id >= sizeof(ctx->bufs)/sizeof(ctx->bufs[0])) {
		vxd_err("  VXD UNMAP FAIL: buffer id (%d) too large!\n", buf_id);
		ret = 1;
		goto out;
	}

	if (ctx->bufs[buf_id].id == VXD_INVALID_ID) {
		vxd_err("  VXD UNMAP FAIL: invalid buffer id!\n");
		ret = 1;
		goto out;
	}

	vxd_dbg("  VXD UNMAP: driver buf id: %d, size: %zu, virt: 0x%08x\n",
			ctx->bufs[buf_id].id, ctx->bufs[buf_id].size,
			ctx->bufs[buf_id].virt_addr);

	data.stream_id = str_id;
	data.buf_id = ctx->bufs[buf_id].id;

	ret = vxd_ioctl(ctx->fd, VXD_IOCTL_VXD_UNMAP, (void *)&data);
	if (ret) {
		vxd_err("  VXD UNMAP IOCTL FAIL: %d\n", ret);
		goto out;
	}

	ctx->bufs[buf_id].virt_addr = 0;

out:
	free(params);
	free(none);

	if (ret)
		vxd_err("  VXD UNMAP FAIL: ret: %d\n", ret);
	else
		vxd_dbg("  VXD UNMAP OK\n");

	return ret;
}

/* MAP THE BUFFER INTO PROCESS ADDRESS SPACE */
void vxd_action_mmap_help(const char *name)
{
	printf("\tMmap a buffer into process address space.\n"
		"\t  --%s=<buf_id>\n"
		"\tE.g. to map the first allocated buffer:\n"
		"\t  --%s=1\n", name, name);
}

int vxd_action_mmap_exec(const char * arg)
{
	struct vxd_util_ctx *ctx = &vxd_util_ctxs[cur_vxd_ctx];
	int ret = 0, *iptr;
	const char *args = strchr(arg, ' '); /* skip action code */
	uint32_t *params, buf_id;
	uint8_t *none;
	size_t params_count = vxd_parse_ints(args, 0, &params, &none);
	void *ptr;

	vxd_dbg("  MMAP: ctx: %d, fd: %d\n", cur_vxd_ctx, ctx->fd);

	if (params_count < 1) {
		vxd_err("  MMAP FAIL: buffer id missing!\n");
		ret = 1;
		goto out;
	}

	buf_id = params[0];
	vxd_dbg("  MMAP: buf id: %d\n", buf_id);

	if (buf_id >= sizeof(ctx->bufs)/sizeof(ctx->bufs[0])) {
		vxd_err("  VXD MAP FAIL: buffer id (%d) too large!\n", buf_id);
		ret = 1;
		goto out;
	}

	if (ctx->bufs[buf_id].id == VXD_INVALID_ID) {
		vxd_err("  MMAP FAIL: invalid buffer id!\n");
		ret = 1;
		goto out;
	}

	vxd_dbg("  MMAP: driver buf id: %d, size: %zu\n", ctx->bufs[buf_id].id,
			ctx->bufs[buf_id].size);


	ptr = mmap(NULL, ctx->bufs[buf_id].size,
			PROT_READ | PROT_WRITE, MAP_SHARED, ctx->fd,
			getpagesize() * ctx->bufs[buf_id].id);
	if (ptr == MAP_FAILED) {
		vxd_err("  MMAP: mmap failed!\n");
		ret = 1;
		goto out;
	}

	vxd_dbg("  MMAP: mapped to: %p\n", ptr);
	iptr = ptr;
	*iptr = 0xdeadbeef;

	ctx->bufs[buf_id].ptr = ptr;

out:
	free(params);
	free(none);

	if (ret)
		vxd_err("  MMAP FAIL: ret: %d\n", ret);
	else
		vxd_dbg("  MMAP OK\n");

	return ret;
}

/* UNMAP THE BUFFER FROM PROCESS ADDRESS SPACE */
void vxd_action_munmap_help(const char *name)
{
	printf("\tUnmap the buffer from process address space.\n"
		"\t  --%s=<buf_id>\n"
		"\tE.g. to unmap the first allocated buffer:\n"
		"\t  --%s=1\n", name, name);
}

int vxd_action_munmap_exec(const char * arg)
{
	struct vxd_util_ctx *ctx = &vxd_util_ctxs[cur_vxd_ctx];
	int ret = 0;
	const char *args = strchr(arg, ' '); /* skip action code */
	uint32_t *params, buf_id;
	uint8_t *none;
	size_t params_count = vxd_parse_ints(args, 0, &params, &none);

	vxd_dbg("  MUNMAP: ctx: %d, fd: %d\n", cur_vxd_ctx, ctx->fd);

	if (params_count < 1) {
		vxd_err("  MUNMAP FAIL: buffer id missing!\n");
		ret = 1;
		goto out;
	}

	buf_id = params[0];
	vxd_dbg("  MUNMAP: buf id: %d\n", buf_id);

	if (buf_id >= sizeof(ctx->bufs)/sizeof(ctx->bufs[0])) {
		vxd_err("  VXD MAP FAIL: buffer id (%d) too large!\n", buf_id);
		ret = 1;
		goto out;
	}

	if (ctx->bufs[buf_id].id == VXD_INVALID_ID) {
		vxd_err("  MNUMAP FAIL: invalid buffer id!\n");
		ret = 1;
		goto out;
	}

	vxd_dbg("  MUNMAP: driver buf id: %d, size: %zu, virt: 0x%08x, cpu: %p\n",
			ctx->bufs[buf_id].id, ctx->bufs[buf_id].size,
			ctx->bufs[buf_id].virt_addr, ctx->bufs[buf_id].ptr);

	ret = munmap(ctx->bufs[buf_id].ptr, ctx->bufs[buf_id].size);
	if (ret != 0) {
		vxd_err("  MUNMAP: munmap failed! (%d)\n", ret);
		ret = 1;
		goto out;
	}

	ctx->bufs[buf_id].ptr = NULL;
out:
	free(params);
	free(none);

	if (ret)
		vxd_err("  MUNMAP FAIL: ret: %d\n", ret);
	else
		vxd_dbg("  MUNMAP OK\n");

	return ret;
}

/* POPULATE MMAPED BUFFER WITH COINTENT OF A FILE */
void vxd_action_fill_buf_help(const char *name)
{
	printf("\tRead the content of a file and copy it to the specified buffer.\n"
		"\t  --%s=<buf_id> <number of bytes to read> <file_name>\n"
		"\tE.g. to copy 100 bytes from file image.bin to the second allocated buffer:\n"
		"\t  --%s=\"2 100 image.bin\"\n", name, name);
}

int vxd_action_fill_buf_exec(const char * arg)
{
	struct vxd_util_ctx *ctx = &vxd_util_ctxs[cur_vxd_ctx];
	FILE *file = NULL;
	const char *args = strchr(arg, ' '), *file_name; /* skip action code */
	uint32_t *params, buf_id, to_copy, ret;
	uint8_t *none;
	size_t params_count = vxd_parse_ints(args, 0, &params, &none);
	void *ptr;

	vxd_dbg("  FILL BUF: ctx: %d, fd: %d\n", cur_vxd_ctx, ctx->fd);

	if (params_count < 1) {
		vxd_err("  FILL BUF FAIL: buffer id missing!\n");
		ret = 1;
		goto out;
	}

	if (params_count < 2) {
		vxd_err("  FILL BUF FAIL: size missing!\n");
		ret = 1;
		goto out;
	}

	if (params_count < 3) {
		vxd_err("  FILL BUF FAIL: file name missing!\n");
		ret = 1;
		goto out;
	}

	buf_id = params[0];
	to_copy = params[1];
	vxd_dbg("  FILL BUF: buf id: %d\n", buf_id);

	if (buf_id >= sizeof(ctx->bufs)/sizeof(ctx->bufs[0])) {
		vxd_err("  VXD MAP FAIL: buffer id (%d) too large!\n", buf_id);
		ret = 1;
		goto out;
	}

	if (ctx->bufs[buf_id].id == VXD_INVALID_ID) {
		vxd_err("  FILL BUF FAIL: invalid buffer id!\n");
		ret = 1;
		goto out;
	}

	if (ctx->bufs[buf_id].ptr == NULL) {
		vxd_err("  FILL BUF FAIL: buffer not mapped!\n");
		ret = 1;
		goto out;
	}

	if (ctx->bufs[buf_id].size < to_copy) {
		vxd_err("  FILL BUF FAIL: mapped buffer too small!\n");
		ret = 1;
		goto out;
	}

	vxd_dbg("  FILL BUF: driver buf id: %d, size: %zu\n", ctx->bufs[buf_id].id,
			ctx->bufs[buf_id].size);

	file_name = vxd_get_arg_no(3, arg);
	vxd_dbg("  FILL BUF: file name: %s\n", file_name);

	file = fopen(file_name, "r");
	if (!file) {
		perror("  FILL BUF: fopen: ");
		vxd_dbg("  FILL BUF: failed to open file!\n");
		ret = 1;
		goto out;
	}

	ret = fread(ctx->bufs[buf_id].ptr, to_copy, 1, file);
	if (ret != 1) {
		vxd_dbg("  FILL BUF: fread failed: %d!\n", ret);
		ret = 1;
		goto out;
	}

	ret = 0;

out:
	if (file)
		fclose(file);
	free(params);
	free(none);

	if (ret)
		vxd_err("  FILL BUF FAIL: ret: %d\n", ret);
	else
		vxd_dbg("  FILL BUF OK\n");

	return ret;
}


/* DUMP CONTENT OF A BUFFER INTO A FILE */
void vxd_action_dump_buf_help(const char *name)
{
	printf("\tSave content of a buffer into the file or print to stdout\n"
		"\t  --%s=<buf_id> <file_name>\n"
		"\tIf <file_name> is set to stdout, content of buffer is dumped "
		"to the console.\n"
		"\tE.g. to save content of the fourth allocated buffer to image.bin:\n"
		"\t  --%s=\"4 image.bin\"\n", name, name);
}

int vxd_action_dump_buf_exec(const char * arg)
{
	struct vxd_util_ctx *ctx = &vxd_util_ctxs[cur_vxd_ctx];
	FILE *file = NULL;
	const char *args = strchr(arg, ' '), *file_name; /* skip action code */
	uint32_t *params, buf_id, to_copy, ret;
	uint8_t *none;
	size_t params_count = vxd_parse_ints(args, 0, &params, &none);
	void *ptr;

	vxd_dbg("  DUMP BUF: ctx: %d, fd: %d\n", cur_vxd_ctx, ctx->fd);

	if (params_count < 1) {
		vxd_err("  DUMP BUF FAIL: buffer id missing!\n");
		ret = 1;
		goto out;
	}

	if (params_count < 2) {
		vxd_err("  DUMP BUF FAIL: file name missing!\n");
		ret = 1;
		goto out;
	}

	buf_id = params[0];
	vxd_dbg("  DUMP BUF: buf id: %d\n", buf_id);

	if (buf_id >= sizeof(ctx->bufs)/sizeof(ctx->bufs[0])) {
		vxd_err("  VXD MAP FAIL: buffer id (%d) too large!\n", buf_id);
		ret = 1;
		goto out;
	}

	if (ctx->bufs[buf_id].id == VXD_INVALID_ID) {
		vxd_err("  DUMP BUF FAIL: invalid buffer id!\n");
		ret = 1;
		goto out;
	}

	if (ctx->bufs[buf_id].ptr == NULL) {
		vxd_err("  DUMP BUF FAIL: buffer not mapped!\n");
		ret = 1;
		goto out;
	}

	vxd_dbg("  DUMP BUF: driver buf id: %d, size: %zu\n", ctx->bufs[buf_id].id,
			ctx->bufs[buf_id].size);

	file_name = vxd_get_arg_no(2, arg);
	vxd_dbg("  DUMP BUF: file name: %s\n", file_name);

	if (!strcmp(file_name, "stdout")) {
		uint32_t val[4], j, i = 0, *buf = (uint32_t *)ctx->bufs[buf_id].ptr;
		uint32_t size_wrds = ctx->bufs[buf_id].size/sizeof(uint32_t);
		while (i < size_wrds) {
			uint32_t start_off = i;
			for (j = 0; j < sizeof(val)/sizeof(val[0]); j++) {
				val[j] = (i < size_wrds) ? buf[i] : 0;
				i++;
			}
			fprintf(stdout, "  DUMP BUF: 0x%08x: 0x%08x 0x%08x 0x%08x 0x%08x\n",
					start_off*sizeof(uint32_t), val[0], val[1], val[2], val[3]);
		}
	} else {
		file = fopen(file_name, "w");
		if (!file) {
			perror("  DUMP BUF: fopen: ");
			vxd_dbg("  DUMP BUF: failed to open file!\n");
			ret = 1;
			goto out;
		}

		ret = fwrite(ctx->bufs[buf_id].ptr, ctx->bufs[buf_id].size, 1, file);
		if (ret != 1) {
			vxd_dbg("  DUMP BUF: fwrite failed: %d!\n", ret);
			ret = 1;
			goto out;
		}
	}

	ret = 0;

out:
	if (file)
		fclose(file);
	free(params);
	free(none);

	if (ret)
		vxd_err("  DUMP BUF FAIL: ret: %d\n", ret);
	else
		vxd_dbg("  DUMP BUF OK\n");

	return ret;
}


/* POPULATE MMAPED BUFFER WITH A PATTERN */
void vxd_action_popu_buf_help(const char *name)
{
	printf("\tPopulate the buffer with a pattern.\n"
		"\t  --%s=<buf_id> <pattern_type>\n"
		"\tAvailable pattern types are %d (zero the buffer) and %d"
		"(increasing 32 bit unsigned ints, binary form).\n",
		name, PATT_TYPE_ZERO, PATT_TYPE_INC_UINT);
}


int vxd_action_popu_buf_exec(const char * arg)
{
	struct vxd_util_ctx *ctx = &vxd_util_ctxs[cur_vxd_ctx];
	const char *args = strchr(arg, ' '); /* skip action code */
	uint32_t *params, buf_id, ret, *ptr, pat_type, i;
	uint8_t *none;
	size_t params_count = vxd_parse_ints(args, 0, &params, &none);

	vxd_dbg("  POPU BUF: ctx: %d, fd: %d\n", cur_vxd_ctx, ctx->fd);

	if (params_count < 1) {
		vxd_err("  POPU BUF FAIL: buffer id missing!\n");
		ret = 1;
		goto out;
	}

	if (params_count < 2) {
		vxd_err("  POPU BUF FAIL: pattern type missing!\n");
		ret = 1;
		goto out;
	}

	buf_id = params[0];
	pat_type = params[1];
	vxd_dbg("  POPU BUF: buf id: %d, pat_type: %d\n", buf_id, pat_type);

	if (buf_id >= sizeof(ctx->bufs)/sizeof(ctx->bufs[0])) {
		vxd_err("  POPU BUF FAIL: buffer id (%d) too large!\n", buf_id);
		ret = 1;
		goto out;
	}

	if (ctx->bufs[buf_id].id == VXD_INVALID_ID) {
		vxd_err("  POPU BUF FAIL: invalid buffer id!\n");
		ret = 1;
		goto out;
	}

	if (ctx->bufs[buf_id].ptr == NULL) {
		vxd_err("  POPU BUF FAIL: buffer not mapped!\n");
		ret = 1;
		goto out;
	}

	vxd_dbg("  POPU BUF: driver buf id: %d, size: %zu\n", ctx->bufs[buf_id].id,
			ctx->bufs[buf_id].size);

	ptr = (uint32_t *)ctx->bufs[buf_id].ptr;

	switch (pat_type) {
		case PATT_TYPE_ZERO:
			memset(ctx->bufs[buf_id].ptr, 0, ctx->bufs[buf_id].size);
			break;
		case PATT_TYPE_INC_UINT:
			for (i = 0; i < ctx->bufs[buf_id].size/sizeof(uint32_t); i++) {
				*ptr = i;
				ptr++;
			}
			break;
		default:
			vxd_err("  POPU BUF FAIL: unkown pattern type!\n");
			ret = 1;
			goto out;
	}

	ret = 0;

out:
	free(params);
	free(none);

	if (ret)
		vxd_err("  POPU BUF FAIL: ret: %d\n", ret);
	else
		vxd_dbg("  POPU BUF OK\n");

	return ret;
}

/* VERIFY CONTENT OF A MMAPED BUFFER */
void vxd_action_verif_buf_help(const char *name)
{
	printf("\tVerify content of the buffer against specified pattern.\n"
		"\t  --%s=<buf_id> <pattern_type>\n"
		"\tAvailable pattern types are %d (buffer filled with 0s) and %d"
		"(buffer filled with increasing 32 bit unsigned ints, binary form).\n",
		name, PATT_TYPE_ZERO, PATT_TYPE_INC_UINT);
}


int vxd_action_verif_buf_exec(const char * arg)
{
	struct vxd_util_ctx *ctx = &vxd_util_ctxs[cur_vxd_ctx];
	const char *args = strchr(arg, ' '); /* skip action code */
	uint32_t *params, buf_id, ret, *ptr, pat_type, i;
	uint8_t *none;
	size_t params_count = vxd_parse_ints(args, 0, &params, &none);

	vxd_dbg("  VERIF BUF: ctx: %d, fd: %d\n", cur_vxd_ctx, ctx->fd);

	if (params_count < 1) {
		vxd_err("  VERIF BUF FAIL: buffer id missing!\n");
		ret = 1;
		goto out;
	}

	if (params_count < 2) {
		vxd_err("  VERIF BUF FAIL: pattern type missing!\n");
		ret = 1;
		goto out;
	}

	buf_id = params[0];
	pat_type = params[1];
	vxd_dbg("  VERIF BUF: buf id: %d, pat_type: %d\n", buf_id, pat_type);

	if (buf_id >= sizeof(ctx->bufs)/sizeof(ctx->bufs[0])) {
		vxd_err("  VERIF FAIL: buffer id (%d) too large!\n", buf_id);
		ret = 1;
		goto out;
	}

	if (ctx->bufs[buf_id].id == VXD_INVALID_ID) {
		vxd_err("  VERIF BUF FAIL: invalid buffer id!\n");
		ret = 1;
		goto out;
	}

	if (ctx->bufs[buf_id].ptr == NULL) {
		vxd_err("  VERIF BUF FAIL: buffer not mapped!\n");
		ret = 1;
		goto out;
	}

	vxd_dbg("  VERIF BUF: driver buf id: %d, size: %zu\n", ctx->bufs[buf_id].id,
			ctx->bufs[buf_id].size);

	ptr = (uint32_t *)ctx->bufs[buf_id].ptr;

	switch (pat_type) {
		case PATT_TYPE_ZERO:
			for (i = 0; i < ctx->bufs[buf_id].size/sizeof(uint32_t); i++) {
				if (*ptr != 0) {
					vxd_err("  VERIF BUF FAIL: mismatch @ %p (%u)!\n", ptr, i);
					ret = 1;
					goto out;
				}
				ptr++;
			}
			break;
		case PATT_TYPE_INC_UINT:
			for (i = 0; i < ctx->bufs[buf_id].size/sizeof(uint32_t); i++) {
				if (*ptr != i) {
					vxd_err("  VERIF BUF FAIL: mismatch @ %p (%u)!\n", ptr, i);
					ret = 1;
					goto out;
				}
				ptr++;
			}
			break;
		default:
			vxd_err("  VERIF BUF FAIL: unkown pattern type!\n");
			ret = 1;
			goto out;
	}

	ret = 0;

out:
	free(params);
	free(none);

	if (ret)
		vxd_err("  VERIF BUF FAIL: ret: %d\n", ret);
	else
		vxd_dbg("  VERIF BUF OK\n");

	return ret;
}
