/*!
 *****************************************************************************
 *
 * @File       helpers.c
 * ---------------------------------------------------------------------------
 *
 * Copyright (c) Imagination Technologies Ltd.
 * 
 * The contents of this file are subject to the MIT license as set out below.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
 * THE SOFTWARE.
 * 
 * Alternatively, the contents of this file may be used under the terms of the 
 * GNU General Public License Version 2 ("GPL")in which case the provisions of
 * GPL are applicable instead of those above. 
 * 
 * If you wish to allow use of your version of this file only under the terms 
 * of GPL, and not to allow others to use your version of this file under the 
 * terms of the MIT license, indicate your decision by deleting the provisions 
 * above and replace them with the notice and other provisions required by GPL 
 * as set out in the file called "GPLHEADER" included in this distribution. If 
 * you do not delete the provisions above, a recipient may use your version of 
 * this file under the terms of either the MIT license or GPL.
 * 
 * This License is also included in this distribution in the file called 
 * "MIT_COPYING".
 *
 *****************************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <uapi/vxd.h>

#include "vxd_util.h"
#include "helpers.h"

#define VXD_DEV_NAME_LEN 16
int vxd_open(int dev_id)
{
	char name[VXD_DEV_NAME_LEN];
	int fd, ret;

	if (snprintf(name, VXD_DEV_NAME_LEN, VXD_DEV_PATH, dev_id) < 0)
		return -1;

	ret = open(name, O_RDWR);
	if (ret < 0)
		perror("  open");

	return ret;
}

int vxd_close(int fd)
{
	int ret = close(fd);
	if (ret)
		perror("  close");
	return ret;
}

int vxd_ioctl(int fd, unsigned int code, void *data)
{
	long stream_id;
	int ret;
	ret = ioctl(fd, code, data);
	if (ret)
		perror("  ioctl");

	return ret;
}

int vxd_send_msg(int fd, uint32_t *buf, size_t count)
{
	unsigned int ret;
	size_t count_bytes = count*sizeof(uint32_t);

	ret = write(fd, buf, count_bytes);
	if (ret == count_bytes) {
		return 0;
	} else {
		perror("  write");
		return ret;
	}
}

int vxd_recv_msg(int fd, uint32_t *buf, size_t count)
{

	return read(fd, buf, count*sizeof(uint32_t));
}

int vxd_parse_ints(const char *arg, size_t alloc_prefix,
		uint32_t **res, uint8_t **none)
{
	char *tmp = NULL;
	const char *next, *prev;
	const char *delim = " ";
	int num_items = 0;
	int i = 0;
	size_t str_len;
	char *str;
	int array_len = 0;

	while (*arg == ' ') arg++;
	str_len = strlen(arg) + 1;
	str = malloc(str_len);

	if (!str) {
		vxd_err("failed to alloc string!\n");
		abort();
	}

	/* count number of arguments */
	memcpy(str, arg, str_len);
	strtok_r(str, delim, &tmp);
	next = str;
	while (next) {
		prev = next;
		next = strtok_r(NULL, delim, &tmp);
#if 0
		vxd_dbg("next: %p -- %s, prev: %p -- %s, num_items: %d\n",
				next, next, prev, prev, num_items);
#endif
		if (strlen(prev) > 0) num_items++;
	}

	array_len = alloc_prefix + num_items;

	(*res) = malloc(array_len * sizeof(uint32_t));
	if (!(*res)) {
		vxd_err("Failed to alloc ints array!\n");
		abort();
	}
	(*none) = malloc(array_len * sizeof(uint8_t));
	if (!(*none)) {
		vxd_err("Failed to alloc none array!\n");
		abort();
	}
	memset(*none, 0, array_len * sizeof(uint8_t));

	/* parse arguments */
	memcpy(str, arg, str_len);
	next = str;
	tmp = NULL;
	strtok_r(str, delim, &tmp);
	i = alloc_prefix;
	while (next && i < array_len) {
		prev = next;
		next = strtok_r(NULL, delim, &tmp);
		if (strlen(prev))
			(*res)[i] = strtol(prev, NULL, 16);
		if (!strcmp(prev, VXD_NONE_LABEL))
			(*none)[i] = 1;
		else
			(*none)[i] = 0;
		i++;
	}

	free(str);
	return num_items;
}

const char* vxd_get_arg_no(int arg_no, const char *src)
{
	/* skip leading spaces */
	while (*src == ' ' && *src != 0) src++;

	while (arg_no > 0 && *src != 0) {
		/* skip argument */
		while (*src != ' ' && *src != 0) src++;
		/* skip trailing spaces */
		while (*src == ' ' && *src != 0) src++;
		arg_no--;
	}

	return src;
}

