/*!
 *****************************************************************************
 *
 * @File       main.c
 * ---------------------------------------------------------------------------
 *
 * Copyright (c) Imagination Technologies Ltd.
 *
 * The contents of this file are subject to the MIT license as set out below.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the terms of the
 * GNU General Public License Version 2 ("GPL")in which case the provisions of
 * GPL are applicable instead of those above.
 *
 * If you wish to allow use of your version of this file only under the terms
 * of GPL, and not to allow others to use your version of this file under the
 * terms of the MIT license, indicate your decision by deleting the provisions
 * above and replace them with the notice and other provisions required by GPL
 * as set out in the file called "GPLHEADER" included in this distribution. If
 * you do not delete the provisions above, a recipient may use your version of
 * this file under the terms of either the MIT license or GPL.
 *
 * This License is also included in this distribution in the file called
 * "MIT_COPYING".
 *
 *****************************************************************************/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <stdint.h>
#include <string.h>
#include <fcntl.h>

#include "vxd_util.h"

#define VXD_ACTION_OPT_NAME(NAME) VXD_ACTION_##NAME##_OPT_NAME
#define VXD_ACTION_ID(NAME) VXD_ACTION_##NAME##_ID


#define VXD_ACTION_BINDING(NAME, NO) \
	static const int (VXD_ACTION_ID(NAME)) = NO; \
	static const char *(VXD_ACTION_OPT_NAME(NAME)) = #NAME; \

#define VXD_ACTION(NAME) \
	.no = VXD_ACTION_ID(NAME), \
	.name = VXD_ACTION_OPT_NAME(NAME), \
	.help_fn = vxd_action_##NAME##_help, \
	.exec_fn = vxd_action_##NAME##_exec,

VXD_ACTION_BINDING(open, 1);
VXD_ACTION_BINDING(close, 2);
VXD_ACTION_BINDING(get_props, 3);
VXD_ACTION_BINDING(str_creat, 10);
VXD_ACTION_BINDING(str_destr, 11);
VXD_ACTION_BINDING(str_switch, 12);
VXD_ACTION_BINDING(send_msg, 20);
VXD_ACTION_BINDING(recv_msg, 21);
VXD_ACTION_BINDING(alloc, 30);
VXD_ACTION_BINDING(free, 31);
VXD_ACTION_BINDING(vxd_map, 40);
VXD_ACTION_BINDING(vxd_unmap, 41);
VXD_ACTION_BINDING(mmap, 50);
VXD_ACTION_BINDING(munmap, 51);
VXD_ACTION_BINDING(fill_buf, 60);
VXD_ACTION_BINDING(dump_buf, 61);
VXD_ACTION_BINDING(popu_buf, 62);
VXD_ACTION_BINDING(verif_buf, 63);

static struct vxd_action * vxd_actions;
static int num_vxd_actions;

static void vxd_init_actions()
{
	struct vxd_action defined_actions[] = {
		{ VXD_ACTION(open) },
		{ VXD_ACTION(close) },
		{ VXD_ACTION(get_props) },
		{ VXD_ACTION(str_creat) },
		{ VXD_ACTION(str_destr) },
		{ VXD_ACTION(str_switch) },
		{ VXD_ACTION(send_msg) },
		{ VXD_ACTION(recv_msg) },
		{ VXD_ACTION(alloc) },
		{ VXD_ACTION(free) },
		{ VXD_ACTION(vxd_map) },
		{ VXD_ACTION(vxd_unmap) },
		{ VXD_ACTION(mmap) },
		{ VXD_ACTION(munmap) },
		{ VXD_ACTION(fill_buf) },
		{ VXD_ACTION(dump_buf) },
		{ VXD_ACTION(popu_buf) },
		{ VXD_ACTION(verif_buf) },
	};
	vxd_actions = malloc(sizeof(defined_actions));
	if (!vxd_actions) {
		vxd_err("Failed to allocate actions!\n");
		abort();
	}
	memcpy(vxd_actions, defined_actions, sizeof(defined_actions));
	num_vxd_actions = sizeof(defined_actions)/sizeof(struct vxd_action);
}

void vxd_help(void)
{
	size_t num = num_vxd_actions;
	unsigned int i;
	printf("Simple utility for testing VXD driver\n\n");
	printf("Example of usage:\n");
	printf("  vxd_util --open --str_creat --send_msg=\"0x0 0x1 0x20\"\n\n");
	printf("Available actions:\n");
	for (i = 0; i < num; i++) {
		printf("  --%s [%03d]:\n", vxd_actions[i].name,
				vxd_actions[i].no);
		if (vxd_actions[i].help_fn)
			vxd_actions[i].help_fn(vxd_actions[i].name);
		printf("\n");
	}
	printf("Number in brackets spcifies action number, can be used");
	printf(" with -a option to shorten the expressions, e.g.:\n");
	printf("  vxd_util -a 1 -a 10 -a 20\" 0x0 0x1 0x20\"\n\n");
	printf("\n\n");
	printf("Available options:\n");
	printf(" --help, -h:\n");
	printf("\tDisplay this help\n\n");
	printf("\tAlternatively, send actions over STDIN. For example:\n");
	printf("\t  echo -e \"1\\n10\\n20 100\\n21 N N N N\" | vxd_util\n");
}

void vxd_action(const char *opt)
{
	int no, ret;
	unsigned int i;
	size_t count = num_vxd_actions;
	printf("\nACTION: %s\n", opt);
	no = strtol(opt, NULL, 10);
	if (no < 1) {
		vxd_err("  Failed to get action number, ignoring\n");
		return;
	}

	for (i = 0; i < count; i++)
		if (vxd_actions[i].no == no)
			break;

	if (i == count) {
		vxd_err("  Action %d not found!\n", no);
		return;
	}

	if (!vxd_actions[i].exec_fn) {
		vxd_err("  Action %d not implemented!\n", no);
		return;
	}

	if (vxd_actions[i].exec_fn(opt)) {
		printf("\nACTION FAILED, aborting!\n");
		abort();
	}

}

#define MAX_OPT_LEN (1024 * 32)

int main(int argc, char **argv)
{
	int c;
	int digit_optind = 0;
	int action_id = -1;

	printf("vxd_util [PID=%d]\n", getpid());
	vxd_init_actions();
	vxd_util_init();

	while (1) {
		int this_option_optind = optind ? optind : 1;
		int option_index = 0, i, id;
		struct option options_base[] = {
			{ "action", required_argument, 0, 'a' },
			{ "help", no_argument, 0, 'h' },
		};

		/* dynamically attach options corresponding to vxd actions */
		struct option * options =
			malloc(sizeof(struct option)*(num_vxd_actions + 1) +
				sizeof(options_base));
		if (!options) {
			vxd_err("Failed to allocate options!\n");
			abort();
		}
		memcpy(options, options_base, sizeof(options_base));
		id = sizeof(options_base)/sizeof(struct option);
		for (i = 0; i < num_vxd_actions; i++) {
			options[id].name = vxd_actions[i].name;
			options[id].has_arg = optional_argument;
			options[id].flag= &action_id;
			options[id].val = vxd_actions[i].no;
			id++;
		}
		/* NULL entries indicate end of options array */
		options[id].name = NULL;
		options[id].has_arg = 0;
		options[id].flag= NULL;
		options[id].val = 0;

		c = getopt_long(argc, argv, "a:hg", options, &option_index);
		if (c == -1) {
			free(options);
			break;
		}

		switch (c) {

		case 0: {
			int res;
			char * buf = malloc(MAX_OPT_LEN);
			if (!buf) {
				vxd_err("Failed to allocate getopt buf!\n");
				abort();
			}
			memset(buf, 0, MAX_OPT_LEN);
			res = snprintf(buf, MAX_OPT_LEN, "%d %s",
					action_id,
					optarg ? optarg : "");
			if (res == MAX_OPT_LEN || res < 0) {
				vxd_err("Failed to process long"
						"option!\n");
				abort();
			}
			vxd_action(buf);
			free(buf);
			break;
		}
		case 'a':
			vxd_action(optarg);
			break;
		case 'h':
			vxd_help();
			break;
		case '?':
			abort();
			break;
		default:
			printf("?? getopt returned character"
					" code 0x%x ??\n", c);
		}

		free(options);
	}

	if(argc == 1) {
		printf("getting commands from stdin. try this: '1\\n10\\n20 100 0 42\\n21 N N N N 42\\n'\n");
		do {
			char buf[100];
			size_t len = 0;
			if(fgets(buf, sizeof(buf), stdin) == NULL)
				break;
			if(buf[0] == '#')
				continue;
			buf[strlen(buf)-1] = ' ';    // vxd_action wants a space after the first parm!
			vxd_action(buf);
		} while(1);
	}
	exit(EXIT_SUCCESS);
}
