/*!
 *****************************************************************************
 *
 * @File       vxd_debugfs_baikal.c
 *
 * Copyright (C) 2020 Baikal Electronics JSC
 *
 * Author: Pavel Parkhomenko <Pavel.Parkhomenko@baikalelectronics.ru>
 *
 * Portions of this file is based on vxd_debugfs.c.
 * Original copyright, licensing terms and conditions follow.
 *
 * ---------------------------------------------------------------------------
 *
 * Copyright (c) Imagination Technologies Ltd.
 *
 * The contents of this file are subject to the MIT license as set out below.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the terms of the
 * GNU General Public License Version 2 ("GPL")in which case the provisions of
 * GPL are applicable instead of those above.
 *
 * If you wish to allow use of your version of this file only under the terms
 * of GPL, and not to allow others to use your version of this file under the
 * terms of the MIT license, indicate your decision by deleting the provisions
 * above and replace them with the notice and other provisions required by GPL
 * as set out in the file called "GPLHEADER" included in this distribution. If
 * you do not delete the provisions above, a recipient may use your version of
 * this file under the terms of either the MIT license or GPL.
 *
 * This License is also included in this distribution in the file called
 * "MIT_COPYING".
 *
 *****************************************************************************/


#include <linux/arm-smccc.h>
#include <linux/debugfs.h>

#include "vxd_common.h"

#define CORE_SMMU_CFG2_NAME "smmu_cfg2"
#define CORE_ARDOMAIN_NAME "ardomain"
#define CORE_AWDOMAIN_NAME "awdomain"
#define CORE_AXCACHE_NAME "axcache"

#define SMMU_CFG2_ARDOMAIN_MASK  GENMASK(13, 12)
#define SMMU_CFG2_ARDOMAIN_SHIFT 12
#define SMMU_CFG2_AWDOMAIN_MASK  GENMASK(5, 4)
#define SMMU_CFG2_AWDOMAIN_SHIFT 4
#define SMMU_CFG2_AXCACHE_MASK   GENMASK(3, 0)
#define SMMU_CFG2_AXCACHE_SHIFT  0

#define BAIKAL_SMC_VDEC_SMMU_SET_CACHE 0x82000300
#define BAIKAL_SMC_VDEC_SMMU_GET_CACHE 0x82000301

static ssize_t vxd_dbgfs_baikal_smmu_read(struct file *file,
		char __user *user_buf, size_t count, loff_t *ppos)
{
	struct arm_smccc_res res;
	char *filename;
	char buf[16];
	int len = 0;

	arm_smccc_smc(BAIKAL_SMC_VDEC_SMMU_GET_CACHE, 0, 0, 0, 0, 0, 0, 0, &res);

	filename = file->f_path.dentry->d_iname;
	if (!strcmp(CORE_SMMU_CFG2_NAME, filename))
		len = snprintf(buf, sizeof(buf), "%lx\n", res.a0);
	else if (!strcmp(CORE_ARDOMAIN_NAME, filename))
		len = snprintf(buf, sizeof(buf), "%lx\n",
			(res.a0 & SMMU_CFG2_ARDOMAIN_MASK) >> SMMU_CFG2_ARDOMAIN_SHIFT);
	else if (!strcmp(CORE_AWDOMAIN_NAME, filename))
		len = snprintf(buf, sizeof(buf), "%lx\n",
			(res.a0 & SMMU_CFG2_AWDOMAIN_MASK) >> SMMU_CFG2_AWDOMAIN_SHIFT);
	else if (!strcmp(CORE_AXCACHE_NAME, filename))
		len = snprintf(buf, sizeof(buf), "%lx\n",
			(res.a0 & SMMU_CFG2_AXCACHE_MASK) >> SMMU_CFG2_AXCACHE_SHIFT);

	return simple_read_from_buffer(user_buf, count, ppos, buf, len);
}

static ssize_t vxd_dbgfs_baikal_smmu_write(struct file *file,
		const char __user *buf, size_t count, loff_t *ppos)
{
	struct arm_smccc_res res;
	char *filename;
	u8 val;
	int ret;

	filename = file->f_path.dentry->d_iname;
	if (!strcmp(CORE_SMMU_CFG2_NAME, filename))
		return -EINVAL;

	ret = kstrtou8_from_user(buf, count, 16, &val);
	if (ret)
		return ret;

	if (!strcmp(CORE_AWDOMAIN_NAME, filename))
		arm_smccc_smc(BAIKAL_SMC_VDEC_SMMU_SET_CACHE, val, -1, -1, 0, 0, 0, 0, &res);
	else if (!strcmp(CORE_ARDOMAIN_NAME, filename))
		arm_smccc_smc(BAIKAL_SMC_VDEC_SMMU_SET_CACHE, -1, val, -1, 0, 0, 0, 0, &res);
	else if (!strcmp(CORE_AXCACHE_NAME, filename))
		arm_smccc_smc(BAIKAL_SMC_VDEC_SMMU_SET_CACHE, -1, -1, val, 0, 0, 0, 0, &res);

	return count;
}

static const struct file_operations vxd_dbgfs_smmu_fops = {
	.owner = THIS_MODULE,
	.open = simple_open,
	.write = vxd_dbgfs_baikal_smmu_write,
	.read = vxd_dbgfs_baikal_smmu_read,
};

int vxd_plat_baikal_dbgfs_populate(struct vxd_dev *vxd, struct dentry *root_dir)
{
	/* SMMU_CFG2 configuration entry */
	if (!debugfs_create_file(CORE_SMMU_CFG2_NAME,
				S_IWUSR, root_dir, vxd,
				&vxd_dbgfs_smmu_fops)) {
		dev_warn(vxd->dev, "%s: failed to create %s dbg file!\n",
				__func__, CORE_SMMU_CFG2_NAME);
		return -ENOENT;
	}

	/* ARDOMAIN configuration entry */
	if (!debugfs_create_file(CORE_ARDOMAIN_NAME,
				S_IWUSR, root_dir, vxd,
				&vxd_dbgfs_smmu_fops)) {
		dev_warn(vxd->dev, "%s: failed to create %s dbg file!\n",
				__func__, CORE_ARDOMAIN_NAME);
		return -ENOENT;
	}

	/* AWDOMAIN configuration entry */
	if (!debugfs_create_file(CORE_AWDOMAIN_NAME,
				S_IWUSR, root_dir, vxd,
				&vxd_dbgfs_smmu_fops)) {
		dev_warn(vxd->dev, "%s: failed to create %s dbg file!\n",
				__func__, CORE_AWDOMAIN_NAME);
		return -ENOENT;
	}

	/* AXCACHE configuration entry */
	if (!debugfs_create_file(CORE_AXCACHE_NAME,
				S_IWUSR, root_dir, vxd,
				&vxd_dbgfs_smmu_fops)) {
		dev_warn(vxd->dev, "%s: failed to create %s dbg file!\n",
				__func__, CORE_AXCACHE_NAME);
		return -ENOENT;
	}

	return 0;
}
